public class MouseCollider extends CollisionObject {
    GCircle circle --> this;
    circle.mat().transparent(true);

    0.1 => float BASE_SIZE;
    BASE_SIZE => radius;
    circle.sca(radius * @(1, 1, 1));

    0.2 => float BASE_BRIGHTNESS;
    0.8 => float MAX_BRIGHTNESS;
    1. => float narrativeBrightnessModifier;

    // Follow mouse ========================================================
    Mouse @mouse;
    false => mouseConnected;
    fun void followMouse(Mouse mouse_) {
        mouse_ @=> mouse;
        true => mouseConnected;
    }

    // Audio Listener ======================================================
    AudioListener audioListener;
    fun void setAudioListener(AudioListener audioListener_) {
        audioListener_ @=> audioListener;
    }
    fun float sizeModifier() {
        return BASE_SIZE + Math.clampf(0.1 * audioListener.getScaledVolume(), 0., 0.02);
    }
    fun float brightnessModifier() {
        return BASE_BRIGHTNESS + Math.clampf(2 * audioListener.getScaledVolume(), 
                                                    0., MAX_BRIGHTNESS - BASE_BRIGHTNESS);
    }

    // Physics engine ======================================================
    10. => float MAX_VELOCITY;
    @(0, 0, 0) => cur_velocity;

    fun vec3 getNewVelocity(float dt, vec3 prev_pos, vec3 new_pos) {
        return (new_pos - prev_pos) / dt;
    }
    fun void capVelocity() {
        Math.min(getNorm(cur_velocity), MAX_VELOCITY) * getUnitVector(cur_velocity) => cur_velocity;
    }

    fun void update(float dt) {
        @(0, 0, 0) => cur_accel;

        mouse.worldPos => vec3 new_pos;
        0. => new_pos.z;
        if (dt > 0) {
            getNewVelocity(dt, cur_pos, new_pos) => cur_velocity;
            capVelocity();
        }
        new_pos => cur_pos;

        // Update visual
        setSize(sizeModifier());
        setPosition(cur_pos);
        setColor(cur_pos, mouse.worldPos, brightnessModifier());
        setBrightness(narrativeBrightnessModifier);
    }


    // Visual stuff ===============================================================
    fun void setSize(float size_) {
        size_ => radius;
        circle.sca(radius * @(1, 1, 1));
    }
    fun void setPosition(vec3 circle_pos) {
        this.pos(circle_pos);
    }
    fun void setColor(vec3 circle_pos, vec3 mouse_pos, float brightness) {
        getPolar(circle_pos - mouse_pos) => polar vec;

        @(Math.map(Math.phase(vec), -Math.PI, Math.PI, 0, 360), Math.mag(vec) / 2., brightness) => vec3 hsv;
        Color.hsv2rgb(hsv) => circle.mat().color;
    }
    fun void setBrightness(float b) {
        circle.mat().alpha(b);
    }

    // Audio stuff ================================================================

    // Math helper functions ======================================================
    fun float getNorm(vec3 vec) {
        return Math.sqrt(vec.x * vec.x + vec.y * vec.y + vec.z * vec.z);
    }
    fun vec3 getUnitVector(vec3 vec) {
        if (getNorm(vec) == 0.) {
            return vec;
        }
        return vec / getNorm(vec);
    }
    fun polar getPolar(vec3 vec) {
        return (vec.x + vec.y * Math.I) $ polar;
    }
}