public class Narrative {
    40 => int NUM_OSC;
    30 => int LOW_NOTE;
    100 => int REQUIRED_SUM_COUNTS_PART1;
    40::second => dur DUR_PART0;
    30::second => dur DUR_PART1;
    20::second => dur DUR_PART2;
    40::second => dur DUR_PART21;
    40::second => dur DUR_PART3;
    30::second => dur DUR_PART4;
    30::second => dur DUR_PART41;


    // External event handlers ===============================
    float noteCounts[NUM_OSC];
    0 => int sumCounts;

    fun void playedNote(int note, int vel, float vol) {
        if (note >= LOW_NOTE && note < LOW_NOTE + NUM_OSC) {
            vol +=> noteCounts[note - LOW_NOTE];
            sumCounts++;
        }
    }

    1. => float INIT_COLLISION_GAIN;
    0.01 => float INIT_REV_MIX;
    1. => float INIT_SLOW_FACTOR;
    0.999 => float SLOW_FACTOR_PART2;
    0.997 => float SLOW_FACTOR_PART3;
    0.96 => float SLOW_FACTOR_PART41;
    INIT_COLLISION_GAIN => float collisionGain;
    INIT_REV_MIX => float collisionRev;
    INIT_SLOW_FACTOR => float collisionSlowFactor;
    -1 => int followMouseToggle;
    -1 => int rubberbandToggle;
    -1 => int narrativeGravity;
    -1 => float mouseColliderBrightness;
    -1 => int mouseColliderCollision;
    -1 => int colorChangeToggle;


    // Some audio stuff ======================================
    Gain gain => PoleZero dcblokewav => LPF l => JCRev rev => dac;
    0.99 => dcblokewav.blockZero;
    0.3 => gain.gain;
    0.02 => rev.mix;
    15000 => l.freq;

    fun void setOutput(UGen output) {
        rev => output;
    }

    SinOsc oscList[NUM_OSC];
    ADSR adsrList[NUM_OSC];

    [0, 2, 4, 5, 7, 9, 11] @=> int majorScale[];
    [majorScale[0], majorScale[2], majorScale[4]] @=> int majorTriad[];
    [0, 3, 7] @=> int minorChord[];
    [0, 3, 6, 10] @=> int dimChord[];


    // The actual core of the experience =====================
    fun void playExperience() {
        initOscs();
        playPart0();
        playPart1();
        playPart2();
        playPart21();
        playPart3();
        playPart4();
        playPart41();
        playPartFinal();
    }
    spork ~ playExperience();

    fun void initOscs() {
        for (0 => int i; i < NUM_OSC; i++) {
            oscList[i] => adsrList[i] => gain;
            adsrList[i].set( 10::ms, 8::ms, 1., 10::ms );
            oscList[i].freq(Math.mtof(i + LOW_NOTE));
            oscList[i].gain(0.);
        }
    }

    fun void playPart0() {
        <<< "Playing part 0" >>>;

        1. => mouseColliderBrightness;

        now + DUR_PART0 => time later;
        while (now < later) {
            GG.nextFrame() => now;
        }
    }
    
    fun void playPart1() {
        <<< "Playing part 1" >>>;
        
        0. => mouseColliderBrightness;

        // Reset counts
        0 => sumCounts;
        for (0 => int i; i < NUM_OSC; i++) {
            0 => noteCounts[i];
        }

        now + DUR_PART1 => time later;
        while (now < later || sumCounts < REQUIRED_SUM_COUNTS_PART1) {
            for (0 => int i; i < NUM_OSC; i++) {
                oscList[i].gain(1. * noteCounts[i] / 10);
                adsrList[i].keyOn();
            }
            GG.nextFrame() => now;
        }
    }
    
    fun void playPart2() {
        <<< "Playing part 2" >>>;

        now + DUR_PART2 => time later;
        while (now < later) {
            (now - later + DUR_PART2) / DUR_PART2 => float progress;

            // Gradually mute collisions and slow them down
            INIT_COLLISION_GAIN * (1 - progress) => collisionGain;
            INIT_REV_MIX + 0.19 * INIT_REV_MIX * progress => collisionRev;
            INIT_SLOW_FACTOR * (1 - progress) + SLOW_FACTOR_PART2 * progress => collisionSlowFactor;
            GG.nextFrame() => now;
        }
        0 => collisionGain;
        0 => collisionRev;
    }

    fun void playPart21() {
        <<< "Playing part 2.1" >>>;

        // Make sure to mute collisions
        0 => collisionGain;
        0 => collisionRev;

        // Stop following mouse
        0 => followMouseToggle;
        0 => rubberbandToggle;

        now + DUR_PART3 => time later;
        while (now < later) {
            // Reset measure
            for (0 => int i; i < NUM_OSC; i++) {
                oscList[i].gain(0.);
                adsrList[i].keyOff();
            }

            // Get key note for measure
            Math.random2(0, 11) => int key;

            // Play each measure
            for (0 => int beat; beat < 4; beat++) {
                key => int key_;
                while (key_ < NUM_OSC) {
                    for (0 => int i; i < majorScale.size(); i++) {
                        key_ + majorScale[i] => int note;
                        if (note < NUM_OSC) {
                            if (Math.random2(0, 2) == 0) {
                                oscList[note].gain(1. * noteCounts[note] / 10);
                                adsrList[note].keyOn();
                            }
                        }
                    }
                    12 +=> key_;
                }

                now + 1.5::second => time next_beat;
                while(now < next_beat) {
                    GG.nextFrame() => now;
                }
            }
        }
    }

    fun void playPart3() {
        <<< "Playing part 3" >>>;

        // Make sure to mute collisions
        0 => collisionGain;
        0 => collisionRev;

        SLOW_FACTOR_PART3 => collisionSlowFactor;

        now + DUR_PART3 => time later;
        while (now < later) {
            // Reset measure
            for (0 => int i; i < NUM_OSC; i++) {
                oscList[i].gain(0.);
                adsrList[i].keyOff();
            }

            // Get key note for measure
            Math.random2(0, 11) => int key;

            // Play each measure
            for (0 => int beat; beat < 4; beat++) {
                key => int key_;
                while (key_ < NUM_OSC) {
                    for (0 => int i; i < majorTriad.size(); i++) {
                        key_ + majorTriad[i] => int note;
                        if (note < NUM_OSC) {
                            if (Math.random2(0, 2) == 0) {
                                oscList[note].gain(1. * noteCounts[note] / 10);
                                adsrList[note].keyOn();
                            }
                        }
                    }
                    12 +=> key_;
                }

                now + 1::second => time next_beat;
                while(now < next_beat) {
                    GG.nextFrame() => now;
                }
            }
        }
    }

    fun void playPart4() {
        <<< "Playing part 4" >>>;

        1. => mouseColliderBrightness;
        1 => mouseColliderCollision;
        0 => colorChangeToggle;

        now + DUR_PART4 => time later;
        while (now < later) {
            // Reset measure
            for (0 => int i; i < NUM_OSC; i++) {
                oscList[i].gain(0.);
                adsrList[i].keyOff();
            }

            // Get key note for measure
            Math.random2(0, 11) => int key;

            // Play each measure
            for (0 => int beat; beat < 4; beat++) {
                key => int key_;
                while (key_ < NUM_OSC) {
                    for (0 => int i; i < minorChord.size(); i++) {
                        key_ + minorChord[i] => int note;
                        if (note < NUM_OSC) {
                            if (Math.random2(0, 2) == 0) {
                                oscList[note].gain(1. * noteCounts[note] / 10);
                                adsrList[note].keyOn();
                            }
                        }
                    }
                    12 +=> key_;
                }

                now + 0.666::second => time next_beat;
                while(now < next_beat) {
                    GG.nextFrame() => now;
                }
            }
        }
    }

    fun void playPart41() {
        <<< "Playing part 4.1" >>>;

        SLOW_FACTOR_PART41 => collisionSlowFactor;

        now + DUR_PART4 => time later;
        while (now < later) {
            // Reset measure
            for (0 => int i; i < NUM_OSC; i++) {
                oscList[i].gain(0.);
                adsrList[i].keyOff();
            }

            // Get key note for measure
            Math.random2(0, 11) => int key;

            // Play each measure
            for (0 => int beat; beat < 4; beat++) {
                key => int key_;
                while (key_ < NUM_OSC) {
                    for (0 => int i; i < dimChord.size(); i++) {
                        key_ + dimChord[i] => int note;
                        if (note < NUM_OSC) {
                            if (Math.random2(0, 2) == 0) {
                                oscList[note].gain(1. * noteCounts[note] / 10);
                                adsrList[note].keyOn();
                            }
                        }
                    }
                    12 +=> key_;
                }

                now + 0.666::second => time next_beat;
                while(now < next_beat) {
                    GG.nextFrame() => now;
                }
            }
        }
    }

    fun void playPartFinal() {
        <<< "Playing final part" >>>;

        1 => narrativeGravity;

        while (true) {
            // Reset measure
            for (0 => int i; i < NUM_OSC; i++) {
                oscList[i].gain(0.);
                adsrList[i].keyOff();
            }

            // Get key note for measure
            Math.random2(0, 11) => int key;

            // Play each measure
            for (0 => int beat; beat < 4; beat++) {
                key => int key_;
                while (key_ < NUM_OSC) {
                    for (0 => int i; i < dimChord.size(); i++) {
                        key_ + dimChord[i] => int note;
                        if (note < NUM_OSC) {
                            if (Math.random2(0, 2) == 0) {
                                oscList[note].gain(1. * noteCounts[note] / 10);
                                adsrList[note].keyOn();
                            }
                        }
                    }
                    12 +=> key_;
                }

                now + 0.666::second => time next_beat;
                while(now < next_beat) {
                    GG.nextFrame() => now;
                }
            }
        }
    }
}