// Initialize Mouse Manager ===================================================
Mouse mouse;
spork ~ mouse.start(0);  // start listening for mouse events
spork ~ mouse.selfUpdate(); // start updating mouse position

// Scene setup ================================================================
GG.windowTitle("Circles!");
GG.scene() @=> GScene @ scene;
GG.camera() @=> GCamera @ cam;
//GG.fullscreen();
GG.scene().backgroundColor( @(0,0,0) );
cam.orthographic();  // Orthographic camera mode for 2D scene

global float screenAspect, screenHeight, screenWidth;

// Mouse listener =============================================================

fun void onLeftClickListener() {
    now => time lastClick;
    while (true) {
        mouse.mouseDownEvents[Mouse.LEFT_CLICK] => now;
        onLeftClick();
    }
}
spork ~ onLeftClickListener();

0 => int num_summoned;
8 => int MAX_SUMMONS;

fun void onLeftClick() {
    summonCircle(num_summoned) => int success;
    success +=> num_summoned;
}

// Circles!!! ====================================================================

Narrative narrative;
CollisionHandler collisionHandler --> scene;
collisionHandler.setNarrative(narrative);
MouseFollower mousefollower --> scene;
mousefollower.followMouse(mouse);
AudioListener audioListener;

// MouseCollider ==========
MouseCollider mouseCollider --> scene;
mouseCollider.followMouse(mouse);
mouseCollider.setAudioListener(audioListener);
//collisionHandler.addCollisionObject(mouseCollider);   // uncomment to allow mouse collision
0 => mouseCollider.narrativeBrightnessModifier; // keep invisible at first

// Generate circles =======

Player players[MAX_SUMMONS];
Sndpeek rubbers[MAX_SUMMONS];
SndpeekManager sndpeekManager --> scene;
Rubberband rubberbands[MAX_SUMMONS];

fun int summonCircle(int i) {
    if (i >= players.size() || i >= rubbers.size() || i >= rubberbands.size()) {
        return false;
    }

    players[i] --> scene;
    players[i].followMouse(mouse);
    players[i].setAudioListener(audioListener);

    rubbers[i].setManager(sndpeekManager);

    rubberbands[i] --> scene;
    rubberbands[i].setRubber(rubbers[i]);
    rubberbands[i].setEndpointObjects(mousefollower, players[i]);

    collisionHandler.addCollisionObject(players[i]);

    return true;
}

// Set input UGens ==================================================
adc => Gain input;
0.3 => input.gain;
sndpeekManager.setInput(input);
audioListener.setInput(input);
collisionHandler.setOutput(input);
narrative.setOutput(input);

// Narrative helper =================================================
fun void narrativeParametersListener() {
    while (true) {
        if (narrative.narrativeGravity != -1) {
            toggleGravity(narrative.narrativeGravity);
            -1 => narrative.narrativeGravity;
        }
        if (narrative.rubberbandToggle != -1) {
            toggleRubberband(narrative.rubberbandToggle);
            -1 => narrative.rubberbandToggle;
        }
        if (narrative.mouseColliderBrightness > 0) {
            setMouseColliderBrightness(narrative.mouseColliderBrightness);
            -1 => narrative.mouseColliderBrightness;
        }
        if (narrative.mouseColliderCollision != -1) {
            setMouseColliderCollision(narrative.mouseColliderCollision);
            -1 => narrative.mouseColliderCollision;
        }
        if (narrative.colorChangeToggle != -1) {
            toggleColorChange(narrative.colorChangeToggle);
            -1 => narrative.colorChangeToggle;
        }
        200::ms => now;
    }
}
spork ~ narrativeParametersListener();

fun void toggleGravity(int toggle) {
    for (0 => int i; i < players.size(); i++) {
        @(0, -20, 0) * toggle => players[i].cur_accel;
    }
}
fun void toggleRubberband(int toggle) {
    for (0 => int i; i < rubbers.size(); i++) {
        toggleRubberbandHelper(rubbers[i], toggle * 1.);
    }
}
fun void toggleRubberbandHelper(Sndpeek @rubber, float b) {
    0.3 => float slew;

    now + 2::second => time later;
    while (now < later) {
        (b - rubber.alpha) * slew +=> rubber.alpha;
        100::ms => now;
    }
    b +=> rubber.alpha;
}
fun void setMouseColliderBrightness(float b) {
    spork ~ setMouseColliderBrightnessHelper(b);
}
fun void setMouseColliderBrightnessHelper(float b) {
    0.1 => float slew;

    now + 5::second => time later;
    while (now < later) {
        (b - mouseCollider.narrativeBrightnessModifier) * slew +=> mouseCollider.narrativeBrightnessModifier;
        100::ms => now;
    }
}
fun void setMouseColliderCollision(int toggle) {
    if (toggle) {
        collisionHandler.addCollisionObject(mouseCollider);
    }
}
fun void toggleColorChange(int toggle) {
    for (0 => int i; i < players.size(); i++) {
        toggle => players[i].color_change;
    }
}

// Game loop ========================================================

while (true) {
    GG.nextFrame() => now;
}