public class SndpeekManager extends GGen {
    // window size
    512 => int WINDOW_SIZE;

    // accumulate samples from mic
    Flip accum => blackhole;
    // set size of flip
    WINDOW_SIZE => accum.size;
    // get a reference for our window for visual tapering of the waveform
    Windowing.hann(WINDOW_SIZE) @=> float window[];

    // sample array
    float samples[0];
    vec3 positions[WINDOW_SIZE];

    fun void setInput(UGen input) {
        input => accum;
    }

    // map audio buffer to 3D positions
    fun void map2waveform( float in[], vec3 out[] )
    {
        if( in.size() != out.size() )
        {
            <<< "size mismatch in map2waveform()", "" >>>;
            return;
        }
        
        // mapping to xyz coordinate
        int i;
        for( auto s : in )
        {
            // space evenly in X
            1./WINDOW_SIZE*i => out[i].x;
            // map y, using window function to taper the ends
            3 * s * window[i] => out[i].y;
            // a constant Z of 0
            0 => out[i].z;
            // increment
            i++;
        }
    }

    // do audio stuff
    fun void doAudio()
    {
        // upchuck to process accum
        accum.upchuck();
        // get the last window size samples (waveform)
        accum.output( samples );
    }

    // graphics render loop
    fun void update(float dt) {
        doAudio();
        map2waveform( samples, positions );
    }
}